"""
Models for describing reproducible computations and graph connections.
"""

from __future__ import annotations

import importlib
import inspect
import traceback
import types
import typing
from collections.abc import Callable
from types import UnionType
from typing import Any, Literal

from derzug.utils.docstring import ParsedDocEntry, parse_numpy_docstring
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

_EMPTY = inspect.Signature.empty


class DocumentedIssue(BaseModel):
    """One warning/exception documented in a callable docstring."""

    name: str
    description: str = ""


class ParameterSpec(BaseModel):
    """One user-configurable computation parameter."""

    name: str
    annotation: str = "object"
    required: bool = True
    default: Any = None
    description: str = ""


class PortSpec(BaseModel):
    """One named graph input or output port."""

    name: str
    annotation: str = "object"
    required: bool = True
    description: str = ""


class ComputationResult(BaseModel):
    """One runtime result produced by invoking a computation."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ok: bool
    value: Any = None
    message: str | None = None
    exception_type: str | None = None
    exception_module: str | None = None
    traceback_text: str | None = None
    error: BaseException | None = Field(default=None, exclude=True)

    @classmethod
    def from_value(cls, value: Any) -> ComputationResult:
        """Build one successful result from a raw callable return value."""
        return cls(ok=True, value=value)

    @classmethod
    def from_exception(cls, exc: BaseException) -> ComputationResult:
        """Build one failed result from an exception raised during invocation."""
        return cls(
            ok=False,
            message=str(exc),
            exception_type=type(exc).__name__,
            exception_module=type(exc).__module__,
            traceback_text="".join(traceback.format_exception(exc)),
            error=exc,
        )

    def is_ok(self) -> bool:
        """Return True when the computation completed successfully."""
        return self.ok

    def is_err(self) -> bool:
        """Return True when the computation failed."""
        return not self.ok

    def unwrap(self) -> Any:
        """Return the value or raise a runtime error describing the failure."""
        if self.ok:
            return self.value
        detail = self.message or "computation failed"
        raise RuntimeError(detail)


class ComputationSpec(BaseModel):
    """
    Immutable description of a callable computation.

    Functions store their runtime parameters and graph inputs on one callable.
    Callable classes split constructor parameters from runtime inputs.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    code_path: str
    kind: Literal["function", "callable_class"]
    parameters: dict[str, ParameterSpec] = Field(default_factory=dict)
    inputs: dict[str, PortSpec] = Field(default_factory=dict)
    outputs: dict[str, PortSpec] = Field(default_factory=dict)
    documented_exceptions: tuple[DocumentedIssue, ...] = ()
    documented_warnings: tuple[DocumentedIssue, ...] = ()
    parameter_order: tuple[str, ...] = ()
    input_order: tuple[str, ...] = ()
    output_order: tuple[str, ...] = ()
    _callable_obj: Callable[..., Any] | type | None = PrivateAttr(default=None)

    @classmethod
    def from_function(
        cls,
        fn,
        *,
        output_names: tuple[str, ...] | None = None,
        input_names: tuple[str, ...] | None = None,
        parameter_names: tuple[str, ...] | None = None,
        metadata_overrides: dict[str, Any] | None = None,
    ) -> ComputationSpec:
        """Build a computation spec from a Python function."""
        if not callable(fn):
            raise TypeError("fn must be callable")

        overrides = _resolve_function_overrides(
            metadata_overrides=metadata_overrides,
            output_names=output_names,
            input_names=input_names,
            parameter_names=parameter_names,
        )
        signature, type_hints, docs = _inspect_callable(fn)
        parameters, inputs, parameter_order, input_order = _split_function_signature(
            signature=signature,
            type_hints=type_hints,
            parameter_docs=docs.parameters,
            input_names=overrides.input_names,
            parameter_names=overrides.parameter_names,
        )

        outputs, output_order = _outputs_from_return_annotation(
            type_hints.get("return", signature.return_annotation),
            docs.returns,
            output_names=overrides.output_names,
        )

        spec = cls(
            code_path=_code_path_for_object(fn),
            kind="function",
            parameters=parameters,
            inputs=inputs,
            outputs=outputs,
            documented_exceptions=_issues_from_doc_entries(docs.raises),
            documented_warnings=_issues_from_doc_entries(docs.warns),
            parameter_order=tuple(parameter_order),
            input_order=tuple(input_order),
            output_order=tuple(output_order),
        )
        spec._callable_obj = fn
        return spec

    @classmethod
    def from_class(
        cls,
        klass: type,
        *,
        output_names: tuple[str, ...] | None = None,
        metadata_overrides: dict[str, Any] | None = None,
    ) -> ComputationSpec:
        """Build a computation spec from a callable class definition."""
        if not inspect.isclass(klass):
            raise TypeError("klass must be a class")
        if "__call__" not in klass.__dict__:
            raise TypeError("klass must define __call__")

        resolved_output_names = _resolve_output_names(
            metadata_overrides=metadata_overrides,
            output_names=output_names,
        )

        init_sig, init_hints, init_docs = _inspect_callable(klass.__init__)
        parameters, parameter_order = _parameter_specs_from_signature(
            signature=init_sig,
            type_hints=init_hints,
            parameter_docs=init_docs.parameters,
        )

        call_sig, call_hints, call_docs = _inspect_callable(klass.__call__)
        inputs, input_order = _input_specs_from_signature(
            signature=call_sig,
            type_hints=call_hints,
            parameter_docs=call_docs.parameters,
        )

        outputs, output_order = _outputs_from_return_annotation(
            call_hints.get("return", call_sig.return_annotation),
            call_docs.returns,
            output_names=resolved_output_names,
        )

        spec = cls(
            code_path=_code_path_for_object(klass),
            kind="callable_class",
            parameters=parameters,
            inputs=inputs,
            outputs=outputs,
            documented_exceptions=_issues_from_doc_entries(call_docs.raises),
            documented_warnings=_issues_from_doc_entries(call_docs.warns),
            parameter_order=tuple(parameter_order),
            input_order=tuple(input_order),
            output_order=tuple(output_order),
        )
        spec._callable_obj = klass
        return spec

    def snapshot(self) -> ComputationSpec:
        """Return a deep copy suitable for one isolated run."""
        snap = self.model_copy(deep=True)
        snap._callable_obj = self._callable_obj
        return snap

    def get_callable(self) -> Callable[..., Any] | type:
        """Return the bound runtime object, importing it on demand if needed."""
        if self._callable_obj is not None:
            return self._callable_obj
        obj = _import_from_code_path(self.code_path)
        self._callable_obj = obj
        return obj

    def invoke(
        self,
        *,
        param_values: dict[str, Any] | None = None,
        input_values: dict[str, Any] | None = None,
    ) -> ComputationResult:
        """Invoke the described computation using current parameter/input values."""
        param_values = param_values or {}
        input_values = input_values or {}

        try:
            _validate_required_values(self.parameters, param_values, kind="parameter")
            _validate_required_values(self.inputs, input_values, kind="input")

            if self.kind == "function":
                return _invoke_function_spec(self, param_values, input_values)

            return _invoke_callable_class_spec(self, param_values, input_values)
        except Exception as exc:
            return ComputationResult.from_exception(exc)


class ComputationState(BaseModel):
    """Mutable runtime values for one computation node."""

    param_values: dict[str, Any] = Field(default_factory=dict)
    input_values: dict[str, Any] = Field(default_factory=dict)
    last_invocation: ComputationResult | None = Field(default=None, exclude=True)

    def snapshot(self) -> ComputationState:
        """Return a deep copy suitable for one isolated run."""
        return self.model_copy(deep=True)


class ComputationNode(BaseModel):
    """One graph node storing computation identity and current values."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    node_id: str
    spec: ComputationSpec
    state: ComputationState = Field(default_factory=ComputationState)

    @classmethod
    def from_function(
        cls,
        node_id: str,
        fn,
        *,
        output_names: tuple[str, ...] | None = None,
        input_names: tuple[str, ...] | None = None,
        parameter_names: tuple[str, ...] | None = None,
        metadata_overrides: dict[str, Any] | None = None,
    ) -> ComputationNode:
        """Convenience constructor from a function."""
        return cls(
            node_id=node_id,
            spec=ComputationSpec.from_function(
                fn,
                output_names=output_names,
                input_names=input_names,
                parameter_names=parameter_names,
                metadata_overrides=metadata_overrides,
            ),
        )

    @classmethod
    def from_class(
        cls,
        node_id: str,
        klass: type,
        *,
        output_names: tuple[str, ...] | None = None,
        metadata_overrides: dict[str, Any] | None = None,
    ) -> ComputationNode:
        """Convenience constructor from a callable class."""
        return cls(
            node_id=node_id,
            spec=ComputationSpec.from_class(
                klass,
                output_names=output_names,
                metadata_overrides=metadata_overrides,
            ),
        )

    def snapshot(self) -> ComputationNode:
        """Return a deep copy suitable for one isolated run."""
        snap = self.model_copy(deep=True)
        snap.spec._callable_obj = self.spec._callable_obj
        return snap

    def invoke(self) -> ComputationResult:
        """Invoke this node and store the latest result/error summary."""
        result = self.spec.invoke(
            param_values=self.state.param_values,
            input_values=self.state.input_values,
        )
        self.state.last_invocation = result
        return result


class Connection(BaseModel):
    """One typed edge between two computation nodes."""

    source_node: str
    source_port: str
    target_node: str
    target_port: str


class _FunctionOverrides(BaseModel):
    """Resolved per-function metadata overrides."""

    output_names: tuple[str, ...] | None = None
    input_names: tuple[str, ...] | None = None
    parameter_names: tuple[str, ...] | None = None


def _resolve_function_overrides(
    *,
    metadata_overrides: dict[str, Any] | None,
    output_names: tuple[str, ...] | None,
    input_names: tuple[str, ...] | None,
    parameter_names: tuple[str, ...] | None,
) -> _FunctionOverrides:
    """Merge direct overrides with metadata overrides for function specs."""
    metadata_overrides = metadata_overrides or {}
    return _FunctionOverrides(
        output_names=output_names or metadata_overrides.get("output_names"),
        input_names=input_names or metadata_overrides.get("input_names"),
        parameter_names=parameter_names or metadata_overrides.get("parameter_names"),
    )


def _resolve_output_names(
    *,
    metadata_overrides: dict[str, Any] | None,
    output_names: tuple[str, ...] | None,
) -> tuple[str, ...] | None:
    """Resolve output names, preferring the explicit method argument."""
    metadata_overrides = metadata_overrides or {}
    return output_names or metadata_overrides.get("output_names")


def _inspect_callable(
    fn: Callable[..., Any],
) -> tuple[inspect.Signature, dict[str, object], Any]:
    """Collect signature, resolved type hints, and parsed docs for one callable."""
    signature = inspect.signature(fn)
    type_hints = _safe_get_type_hints(fn)

    # Parsing once keeps the higher-level constructors focused on model assembly.
    docs = parse_numpy_docstring(inspect.getdoc(fn) or "")
    return signature, type_hints, docs


def _split_function_signature(
    *,
    signature: inspect.Signature,
    type_hints: dict[str, object],
    parameter_docs: dict[str, str],
    input_names: tuple[str, ...] | None,
    parameter_names: tuple[str, ...] | None,
) -> tuple[dict[str, ParameterSpec], dict[str, PortSpec], list[str], list[str]]:
    """Split one function signature into runtime parameters and graph inputs."""
    parameters: dict[str, ParameterSpec] = {}
    inputs: dict[str, PortSpec] = {}
    parameter_order: list[str] = []
    input_order: list[str] = []

    # Name-based overrides should be checked against sets once, not per branch.
    input_name_set = set(input_names or ())
    parameter_name_set = set(parameter_names or ())

    for parameter in signature.parameters.values():
        if _skip_parameter(parameter):
            continue

        if _is_function_input(
            parameter=parameter,
            input_names=input_names,
            parameter_names=parameter_names,
            input_name_set=input_name_set,
            parameter_name_set=parameter_name_set,
        ):
            inputs[parameter.name] = _port_spec_from_parameter(
                parameter=parameter,
                type_hints=type_hints,
                parameter_docs=parameter_docs,
            )
            input_order.append(parameter.name)
            continue

        parameters[parameter.name] = _parameter_spec_from_parameter(
            parameter=parameter,
            type_hints=type_hints,
            parameter_docs=parameter_docs,
        )
        parameter_order.append(parameter.name)

    return parameters, inputs, parameter_order, input_order


def _parameter_specs_from_signature(
    *,
    signature: inspect.Signature,
    type_hints: dict[str, object],
    parameter_docs: dict[str, str],
) -> tuple[dict[str, ParameterSpec], list[str]]:
    """Build constructor parameter specs from one signature."""
    parameters: dict[str, ParameterSpec] = {}
    parameter_order: list[str] = []

    for parameter in signature.parameters.values():
        if _skip_parameter(parameter):
            continue

        parameters[parameter.name] = _parameter_spec_from_parameter(
            parameter=parameter,
            type_hints=type_hints,
            parameter_docs=parameter_docs,
        )
        parameter_order.append(parameter.name)

    return parameters, parameter_order


def _input_specs_from_signature(
    *,
    signature: inspect.Signature,
    type_hints: dict[str, object],
    parameter_docs: dict[str, str],
) -> tuple[dict[str, PortSpec], list[str]]:
    """Build runtime input port specs from one signature."""
    inputs: dict[str, PortSpec] = {}
    input_order: list[str] = []

    for parameter in signature.parameters.values():
        if _skip_parameter(parameter):
            continue

        inputs[parameter.name] = _port_spec_from_parameter(
            parameter=parameter,
            type_hints=type_hints,
            parameter_docs=parameter_docs,
        )
        input_order.append(parameter.name)

    return inputs, input_order


def _skip_parameter(parameter: inspect.Parameter) -> bool:
    """Validate one parameter and skip sentinel receiver arguments."""
    _validate_parameter_kind(parameter)
    return parameter.name == "self"


def _is_function_input(
    *,
    parameter: inspect.Parameter,
    input_names: tuple[str, ...] | None,
    parameter_names: tuple[str, ...] | None,
    input_name_set: set[str],
    parameter_name_set: set[str],
) -> bool:
    """Classify one function argument as a graph input or stored parameter."""
    has_default = parameter.default is not _EMPTY

    # Explicit input names take precedence over all other heuristics.
    if input_names is not None:
        return parameter.name in input_name_set

    # Parameter-name overrides invert the default required-vs-defaulted split.
    if parameter_names is not None:
        return parameter.name not in parameter_name_set

    # Without overrides, required args are runtime inputs and defaults are settings.
    return not has_default


def _parameter_spec_from_parameter(
    *,
    parameter: inspect.Parameter,
    type_hints: dict[str, object],
    parameter_docs: dict[str, str],
) -> ParameterSpec:
    """Convert one inspected parameter into a stored parameter spec."""
    annotation = type_hints.get(parameter.name, parameter.annotation)
    has_default = parameter.default is not _EMPTY
    return ParameterSpec(
        name=parameter.name,
        annotation=_annotation_to_string(annotation),
        required=not has_default,
        default=None if not has_default else parameter.default,
        description=parameter_docs.get(parameter.name, ""),
    )


def _port_spec_from_parameter(
    *,
    parameter: inspect.Parameter,
    type_hints: dict[str, object],
    parameter_docs: dict[str, str],
) -> PortSpec:
    """Convert one inspected parameter into an input/output port spec."""
    annotation = type_hints.get(parameter.name, parameter.annotation)
    return PortSpec(
        name=parameter.name,
        annotation=_annotation_to_string(annotation),
        required=parameter.default is _EMPTY,
        description=parameter_docs.get(parameter.name, ""),
    )


def _invoke_function_spec(
    spec: ComputationSpec,
    param_values: dict[str, Any],
    input_values: dict[str, Any],
) -> ComputationResult:
    """Invoke one function-backed computation spec."""
    fn = spec.get_callable()

    # Preserve parameter ordering first, then let runtime inputs fill the rest.
    kwargs = _kwargs_from_named_specs(spec.parameters, param_values)
    kwargs.update(_kwargs_from_named_specs(spec.inputs, input_values))
    return ComputationResult.from_value(fn(**kwargs))


def _invoke_callable_class_spec(
    spec: ComputationSpec,
    param_values: dict[str, Any],
    input_values: dict[str, Any],
) -> ComputationResult:
    """Invoke one callable-class-backed computation spec."""
    klass = spec.get_callable()

    # Constructor parameters are persisted on the node, inputs arrive at call time.
    init_kwargs = _kwargs_from_named_specs(spec.parameters, param_values)
    call_kwargs = _kwargs_from_named_specs(spec.inputs, input_values)
    instance = klass(**init_kwargs)
    return ComputationResult.from_value(instance(**call_kwargs))


def _safe_get_type_hints(obj) -> dict[str, object]:
    """Return resolved type hints when possible, else raw annotations."""
    try:
        globalns = getattr(obj, "__globals__", None)
        return typing.get_type_hints(obj, globalns=globalns)
    except Exception:
        return dict(getattr(obj, "__annotations__", {}))


def _annotation_to_string(annotation: object) -> str:
    """Return a stable string form for one annotation."""
    if annotation in (_EMPTY, inspect._empty, None):
        return "object"
    if isinstance(annotation, str):
        return annotation
    if getattr(annotation, "__module__", "") == "builtins":
        return annotation.__name__
    text = str(annotation)
    return text.replace("typing.", "")


def _code_path_for_object(obj: object) -> str:
    """Return one stable module/qualname string for a runtime object."""
    module = getattr(obj, "__module__", "__main__")
    qualname = getattr(
        obj,
        "__qualname__",
        getattr(obj, "__name__", type(obj).__name__),
    )
    return f"{module}:{qualname}"


def _import_from_code_path(code_path: str) -> object:
    """Resolve one module:qualname path back into a Python object."""
    module_name, _, qualname = code_path.partition(":")
    if not module_name or not qualname:
        raise ValueError(f"invalid code path {code_path!r}")
    module = importlib.import_module(module_name)
    obj: object = module
    for part in qualname.split("."):
        if part == "<locals>":
            raise ValueError(f"cannot import local object from {code_path!r}")
        obj = getattr(obj, part)
    return obj


def _validate_parameter_kind(parameter: inspect.Parameter) -> None:
    """Reject unsupported callable signature features."""
    if parameter.kind in (
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.VAR_POSITIONAL,
        inspect.Parameter.VAR_KEYWORD,
    ):
        raise ValueError(
            "positional-only parameters, *args, and **kwargs are not supported"
        )


def _outputs_from_return_annotation(
    annotation: object,
    return_entries: tuple[ParsedDocEntry, ...],
    *,
    output_names: tuple[str, ...] | None,
) -> tuple[dict[str, PortSpec], tuple[str, ...]]:
    """Build named output ports from a return annotation and docstring metadata."""
    annotation = _unwrap_optional(annotation)
    doc_names = tuple(entry.name for entry in return_entries if entry.name)
    descriptions_by_name = {
        entry.name: entry.description for entry in return_entries if entry.name
    }

    if _return_is_dict_like(annotation):
        names = output_names or doc_names
        if not names:
            raise ValueError(
                "multi-output dict-like returns require docstring names or output_names"
            )
        value_annotation = _dict_value_annotation(annotation)
        outputs = {
            name: PortSpec(
                name=name,
                annotation=_annotation_to_string(value_annotation),
                required=True,
                description=descriptions_by_name.get(name, ""),
            )
            for name in names
        }
        return outputs, tuple(names)

    tuple_args = _tuple_annotations(annotation)
    if tuple_args is not None:
        names = output_names or doc_names
        if not names:
            raise ValueError(
                "multi-output tuple returns require docstring names or output_names"
            )
        if len(names) != len(tuple_args):
            raise ValueError("output_names must match the number of returned values")
        outputs = {
            name: PortSpec(
                name=name,
                annotation=_annotation_to_string(item_annotation),
                required=True,
                description=descriptions_by_name.get(name, ""),
            )
            for name, item_annotation in zip(names, tuple_args, strict=False)
        }
        return outputs, tuple(names)

    single_name = doc_names[0] if len(doc_names) == 1 else "result"
    description = descriptions_by_name.get(single_name, "")
    if not description and return_entries:
        description = return_entries[0].description
    outputs = {
        single_name: PortSpec(
            name=single_name,
            annotation=_annotation_to_string(annotation),
            required=True,
            description=description,
        )
    }
    return outputs, (single_name,)


def _return_is_dict_like(annotation: object) -> bool:
    """Return True when one return annotation is dict-like."""
    origin = typing.get_origin(annotation)
    return annotation is dict or origin is dict


def _dict_value_annotation(annotation: object) -> object:
    """Return the dict value annotation, defaulting to object."""
    args = typing.get_args(annotation)
    return args[1] if len(args) == 2 else object


def _tuple_annotations(annotation: object) -> tuple[object, ...] | None:
    """Return tuple item annotations for fixed-size tuple returns."""
    origin = typing.get_origin(annotation)
    if annotation is tuple or origin is tuple:
        args = typing.get_args(annotation)
        if not args:
            return None
        if len(args) == 2 and args[1] is Ellipsis:
            return None
        return tuple(args)
    return None


def _unwrap_optional(annotation: object) -> object:
    """Strip Optional / X | None wrappers, returning the inner type."""
    origin = typing.get_origin(annotation)
    if origin in (typing.Union, UnionType, types.UnionType):
        args = [arg for arg in typing.get_args(annotation) if arg is not type(None)]
        if len(args) == 1:
            return args[0]
    return annotation


def _issues_from_doc_entries(
    entries: tuple[ParsedDocEntry, ...],
) -> tuple[DocumentedIssue, ...]:
    """Convert parsed docstring entries into issue models."""
    return tuple(
        DocumentedIssue(name=entry.name, description=entry.description)
        for entry in entries
        if entry.name
    )


def _validate_required_values(
    specs: dict[str, ParameterSpec | PortSpec],
    values: dict[str, Any],
    *,
    kind: str,
) -> None:
    """Raise when one required parameter/input is missing."""
    missing = [
        name for name, spec in specs.items() if spec.required and name not in values
    ]
    if missing:
        joined = ", ".join(missing)
        raise ValueError(f"missing required {kind}(s): {joined}")


def _kwargs_from_named_specs(
    specs: dict[str, ParameterSpec | PortSpec],
    values: dict[str, Any],
) -> dict[str, Any]:
    """Return callable kwargs from currently provided named values."""
    return {name: values[name] for name in specs if name in values}



"""Floating annotation toolbox widget."""

from __future__ import annotations

from pathlib import Path

from AnyQt.QtCore import QPoint, QSize, Qt, Signal
from AnyQt.QtGui import QIcon
from AnyQt.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

_ANNOTATION_ICON_DIR = Path(__file__).resolve().parent / "icons" / "annotations"
_TOOL_METADATA = {
    "point": (
        "Point annotation. Double-click or Shift+click to place. "
        "Double-click an annotation to edit.",
        "point",
    ),
    "line": (
        "Line annotation. Double-click or Shift+click to anchor, then click to finish. "
        "Double-click an annotation to edit.",
        "line",
    ),
    "ellipse": (
        "Ellipse annotation. Double-click or Shift+click to place. "
        "Drag selected ellipses to edit.",
        "ellipse",
    ),
    "hyperbola": (
        "Hyperbola annotation. Double-click or Shift+click to place. "
        "Drag selected hyperbolas to edit.",
        "hyperbola",
    ),
    "box": (
        "Box annotation. Double-click or Shift+click to place. "
        "Double-click an annotation to edit.",
        "box",
    ),
    "delete": ("Delete annotation", "delete"),
}


class AnnotationToolbox(QFrame):
    """Compact floating toolbox for annotation-capable hosts."""

    _BASE_TITLE = "Annotations"

    toolChanged = Signal(str)  # noqa: N815
    hideRequested = Signal()  # noqa: N815
    snapToggled = Signal(bool)  # noqa: N815

    def __init__(
        self, parent: QWidget | None = None, *, tools: tuple[str, ...]
    ) -> None:
        super().__init__(parent)
        self._tools = tools
        self._active_tool: str | None = None
        self._pressed_tool: str | None = None
        self.tool_buttons: dict[str, QToolButton] = {}
        self.hide_button: QToolButton | None = None
        self.snap_button: QToolButton | None = None
        self.header_frame: QFrame | None = None
        self._drag_offset: QPoint | None = None
        self._user_moved = False
        self._build_ui()

    def set_tool(self, tool: str) -> None:
        """Update the checked button to match the active tool."""
        if tool not in self.tool_buttons:
            self.clear_tool()
            return
        for name, button in self.tool_buttons.items():
            button.setChecked(name == tool)
        self._active_tool = tool

    def clear_tool(self) -> None:
        """Clear any checked tool button to leave the toolbox in a neutral state."""
        for button in self.tool_buttons.values():
            button.setChecked(False)
        self._active_tool = None

    def set_dirty(self, dirty: bool) -> None:
        """Reflect unsent annotation changes in the toolbox title."""
        suffix = " *" if dirty else ""
        self.title_label.setText(f"{self._BASE_TITLE}{suffix}")

    def set_snap_enabled(self, enabled: bool) -> None:
        """Update the snap toggle without requiring direct button access."""
        if self.snap_button is not None:
            self.snap_button.setChecked(bool(enabled))

    def snap_enabled(self) -> bool:
        """Return True when the toolbox snap toggle is enabled."""
        return bool(self.snap_button is not None and self.snap_button.isChecked())

    @property
    def user_moved(self) -> bool:
        """Return True after the user drags the toolbox to a custom position."""
        return self._user_moved

    def mousePressEvent(self, event) -> None:
        """Start dragging the toolbox from the title-bar region."""
        if (
            event.button() == Qt.MouseButton.LeftButton
            and self.header_frame is not None
            and self.header_frame.geometry().contains(event.position().toPoint())
        ):
            self._drag_offset = (
                event.globalPosition().toPoint() - self.frameGeometry().topLeft()
            )
            event.accept()
            return
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event) -> None:
        """Drag the toolbox when the left button is held on the frame."""
        if (
            self._drag_offset is not None
            and event.buttons() & Qt.MouseButton.LeftButton
        ):
            self.move(event.globalPosition().toPoint() - self._drag_offset)
            self._user_moved = True
            event.accept()
            return
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event) -> None:
        """Finish any active toolbox drag."""
        self._drag_offset = None
        super().mouseReleaseEvent(event)

    def _build_ui(self) -> None:
        """Create the compact icon-only toolbox."""
        self.setObjectName("annotation-toolbox")
        self.setFrameShape(QFrame.Shape.StyledPanel)
        self.setStyleSheet(
            """
            QFrame#annotation-toolbox {
                background: rgba(246, 246, 246, 242);
                border: 1px solid rgba(62, 62, 62, 170);
                border-radius: 8px;
            }
            QFrame#annotation-toolbox-header {
                background: rgba(228, 228, 228, 245);
                border: 1px solid rgba(0, 0, 0, 0);
                border-bottom: 1px solid rgba(62, 62, 62, 110);
                border-top-left-radius: 6px;
                border-top-right-radius: 6px;
            }
            QLabel#annotation-toolbox-title {
                color: rgb(38, 38, 38);
                font-weight: 600;
                padding-left: 2px;
            }
            QToolButton {
                background: rgba(255, 255, 255, 0);
                border: 1px solid rgba(0, 0, 0, 0);
                border-radius: 4px;
            }
            QToolButton:hover {
                background: rgba(255, 255, 255, 190);
                border: 1px solid rgba(70, 70, 70, 70);
            }
            QToolButton:checked {
                background: rgba(55, 125, 245, 45);
                border: 1px solid rgba(55, 125, 245, 120);
            }
            QToolButton#annotation-toolbox-close {
                color: rgb(80, 80, 80);
                font-weight: 700;
            }
            QToolButton#annotation-toolbox-close:hover {
                background: rgba(220, 70, 70, 55);
                border: 1px solid rgba(180, 40, 40, 90);
            }
            """
        )
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        header_widget = QFrame(self)
        header_widget.setObjectName("annotation-toolbox-header")
        self.header_frame = header_widget
        header = QHBoxLayout(header_widget)
        header.setContentsMargins(8, 5, 5, 5)
        header.setSpacing(4)
        layout.addWidget(header_widget)

        title = QLabel(self._BASE_TITLE, self)
        title.setObjectName("annotation-toolbox-title")
        title.setToolTip("Press S to send annotations")
        header.addWidget(title)
        self.title_label = title

        header.addStretch(1)

        hide_button = QToolButton(self)
        hide_button.setText("X")
        hide_button.setObjectName("annotation-toolbox-close")
        hide_button.setFixedSize(18, 18)
        hide_button.setToolTip("Hide annotation tools")
        hide_button.clicked.connect(self.hideRequested.emit)
        header.addWidget(hide_button)
        self.hide_button = hide_button

        body = QWidget(self)
        body_layout = QVBoxLayout(body)
        body_layout.setContentsMargins(8, 8, 8, 8)
        body_layout.setSpacing(0)
        layout.addWidget(body)

        row = QHBoxLayout()
        row.setContentsMargins(0, 0, 0, 0)
        row.setSpacing(4)
        body_layout.addLayout(row)

        for name in self._tools:
            tooltip, icon_name = _TOOL_METADATA[name]
            button = QToolButton(self)
            button.setCheckable(True)
            button.setFixedSize(24, 24)
            button.setIconSize(QSize(14, 14))
            button.setIcon(QIcon(str(_ANNOTATION_ICON_DIR / f"{icon_name}.svg")))
            button.setToolTip(tooltip)
            button.pressed.connect(
                lambda tool=name: setattr(self, "_pressed_tool", tool)
            )
            button.clicked.connect(
                lambda _checked=False, tool=name: self._on_tool_button_clicked(tool)
            )
            row.addWidget(button)
            self.tool_buttons[name] = button

        snap_button = QToolButton(self)
        snap_button.setText("Snap")
        snap_button.setCheckable(True)
        snap_button.setFixedHeight(24)
        snap_button.setToolTip("Snap point, line, and box edits to nearby annotations")
        snap_button.toggled.connect(self.snapToggled.emit)
        row.addWidget(snap_button)
        self.snap_button = snap_button

        self.hide()
        self.raise_()

    def _on_tool_button_clicked(self, tool: str) -> None:
        """Toggle one tool button on or off and emit the resulting selection."""
        was_active = self._pressed_tool == tool and self._active_tool == tool
        self._pressed_tool = None
        if was_active:
            self.clear_tool()
            self.toolChanged.emit("")
            return
        self.set_tool(tool)
        self.toolChanged.emit(tool)


__all__ = ("AnnotationToolbox",)



"""Shared parsing helpers for textbox values."""

from __future__ import annotations

import re
from typing import Any

import numpy as np
from dascore.units import get_quantity

_NONE_TOKENS: frozenset[str] = frozenset({"", "none", "null"})
_INT_PATTERN = re.compile(r"^[+-]?\d+$")
_FLOAT_PATTERN = re.compile(
    r"^[+-]?(?:\d+\.\d*|\.\d+|\d+(?:[eE][+-]?\d+)|\d+\.\d*[eE][+-]?\d+|\.\d+[eE][+-]?\d+)$"
)
_LEADING_NUMBER_PATTERN = re.compile(
    r"^[\s]*([+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?)"
)


def parse_text_value(text: str) -> int | float | str:
    """
    Parse a textbox value to the narrowest type implied by the input text.

    Integer-shaped text returns ``int``. Explicit float syntax, including
    decimal points and scientific notation, returns ``float``. Everything
    else returns the stripped string unchanged.
    """
    value = text.strip()
    if value == "":
        return ""
    if _INT_PATTERN.fullmatch(value):
        return int(value)
    if _FLOAT_PATTERN.fullmatch(value):
        return float(value)
    return value


def parse_patch_text_value(
    text: str,
    *,
    allow_none: bool = False,
    required: bool = False,
    allow_ellipsis: bool = False,
    allow_quantity: bool = True,
) -> Any | None:
    """
    Parse a textbox value destined for a DASCore patch method.

    This preserves numeric intent from the raw text, while still supporting
    unit-bearing quantities and widget-level sentinels.
    """
    value = text.strip()
    lowered = value.lower()

    if allow_none and lowered in _NONE_TOKENS:
        return None
    if required and lowered in _NONE_TOKENS:
        raise ValueError("value must not be empty")
    if allow_ellipsis and value == "...":
        return Ellipsis

    parsed = parse_text_value(value)
    if not isinstance(parsed, str):
        return parsed
    if not allow_quantity:
        raise ValueError(f"could not parse value {text!r}")

    quantity = get_quantity(value)
    if not getattr(quantity, "dimensionless", False):
        return quantity

    magnitude = quantity.magnitude
    if isinstance(magnitude, int):
        return magnitude
    if isinstance(magnitude, float):
        if _looks_integer_like(value) and magnitude.is_integer():
            return int(magnitude)
        return float(magnitude)

    magnitude_text = str(magnitude)
    magnitude_value = parse_text_value(magnitude_text)
    if isinstance(magnitude_value, int) and not _looks_integer_like(value):
        return float(magnitude)
    return magnitude_value


def parse_timedelta_text(text: str) -> np.timedelta64:
    """Parse free-form timedelta text, including `format_display` output."""
    try:
        return np.timedelta64(text)
    except Exception:
        pass

    parts = text.strip().split(maxsplit=1)
    if len(parts) != 2:
        raise ValueError(f"unsupported timedelta text '{text}'")
    value_text, unit_text = parts
    unit_key = unit_text.strip().lower()
    unit_map = {
        "attosecond": "as",
        "attoseconds": "as",
        "femtosecond": "fs",
        "femtoseconds": "fs",
        "picosecond": "ps",
        "picoseconds": "ps",
        "nanosecond": "ns",
        "nanoseconds": "ns",
        "microsecond": "us",
        "microseconds": "us",
        "millisecond": "ms",
        "milliseconds": "ms",
        "second": "s",
        "seconds": "s",
        "minute": "m",
        "minutes": "m",
        "hour": "h",
        "hours": "h",
        "day": "D",
        "days": "D",
        "week": "W",
        "weeks": "W",
    }
    unit = unit_map.get(unit_key)
    if unit is None:
        raise ValueError(f"unsupported timedelta unit '{unit_text}'")
    return np.timedelta64(int(value_text), unit)


def parse_coord_text_value(text: str, sample: Any, fallback: Any) -> Any:
    """Parse UI text into a coordinate-compatible value using sample dtype."""
    stripped = text.strip()
    if not stripped:
        return fallback

    sample_dtype = getattr(np.asarray(sample), "dtype", None)
    if sample_dtype is not None and np.issubdtype(sample_dtype, np.datetime64):
        return np.datetime64(stripped)
    if sample_dtype is not None and np.issubdtype(sample_dtype, np.timedelta64):
        unit = np.datetime_data(sample_dtype)[0]
        if any(char.isalpha() for char in stripped):
            return parse_timedelta_text(stripped)
        parsed = parse_patch_text_value(stripped, required=True)
        if isinstance(parsed, str):
            return parse_timedelta_text(stripped)
        return np.timedelta64(int(parsed), unit)

    parsed = parse_patch_text_value(stripped, required=True)
    if sample_dtype is not None and np.issubdtype(sample_dtype, np.integer):
        return int(parsed)
    if sample_dtype is not None and np.issubdtype(sample_dtype, np.floating):
        return float(parsed)
    return parsed


def _looks_integer_like(text: str) -> bool:
    """Return True when the leading numeric token is integer-shaped."""
    match = _LEADING_NUMBER_PATTERN.match(text)
    if not match:
        return False
    return _INT_PATTERN.fullmatch(match.group(1)) is not None



"""Generic selection widget for DASCore patch and spool objects."""

from __future__ import annotations

import dascore as dc
from AnyQt.QtCore import QTimer
from Orange.widgets import gui
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.widget import Msg

from derzug.core.zugwidget import ZugWidget
from derzug.models.annotations import AnnotationSet
from derzug.orange import Setting
from derzug.utils.spool import extract_single_patch, filter_contents_by_annotations
from derzug.widgets.selection import SelectionControlsMixin, SelectionMode


class Select(SelectionControlsMixin, ZugWidget):
    """Select subsets of patches or spools using shared left-side controls."""

    name = "Select"
    want_main_area = False
    description = "Select subsets of patches or spools"
    icon = "icons/SelectRows.svg"
    category = "Processing"
    keywords = ("select", "patch", "spool", "subset", "filter")
    priority = 23
    unpack_single_patch = Setting(True)
    saved_selection_basis = Setting("", schema_only=True)
    saved_selection_ranges = Setting([], schema_only=True)
    saved_spool_filters = Setting([], schema_only=True)

    class Error(ZugWidget.Error):
        """Errors shown by the widget."""

        general = Msg("Selection error: {}")

    class Warning(ZugWidget.Warning):
        """Warnings shown by the widget."""

    class Inputs:
        """Input signal definitions."""

        patch = Input("Patch", dc.Patch)
        spool = Input("Spool", dc.BaseSpool)
        annotation_set = Input("Annotations", AnnotationSet)

    class Outputs:
        """Output signal definitions."""

        patch = Output(
            "Patch",
            dc.Patch,
            doc=(
                "If a length 1 spool is emitted, unpack it into a patch and serve "
                "on patch output when unpacking is enabled."
            ),
        )
        spool = Output("Spool", dc.BaseSpool)

    def __init__(self) -> None:
        super().__init__()
        self._init_selection_controls()
        self._prime_saved_selection_state()
        self._patch: dc.Patch | None = None
        self._spool: dc.BaseSpool | None = None
        self._annotation_set: AnnotationSet | None = None
        self._input_kind: str | None = None
        self._preview_selected = None
        self._compact_width_done = False

        params_box = gui.widgetBox(self.controlArea, "Parameters")
        self._build_selection_panel(params_box)
        self.unpack_checkbox = gui.checkBox(
            params_box,
            self,
            "unpack_single_patch",
            "Unpack len1 spool",
            callback=self._emit_selected_output,
        )
        self.unpack_checkbox.setToolTip(self.Outputs.patch.doc or "")
        self._selection_set_status("")

    def showEvent(self, event) -> None:
        """Compact the initial Select window width after the first visible layout."""
        super().showEvent(event)
        if not self._compact_width_done:
            QTimer.singleShot(0, self._compact_initial_width)

    @Inputs.patch
    def set_patch(self, patch: dc.Patch | None) -> None:
        """Receive a patch input and expose patch-range selection controls."""
        self._input_kind = "patch"
        self._patch = patch
        self._spool = None
        self._selection_set_patch_source(patch, notify=False, refresh_ui=False)
        self._emit_selected_output()

    @Inputs.spool
    def set_spool(self, spool: dc.BaseSpool | None) -> None:
        """Receive a spool input and expose metadata selection controls."""
        self._input_kind = "spool"
        self._spool = spool
        self._patch = None
        self._selection_set_spool_source(spool, notify=False, refresh_ui=False)
        self._emit_selected_output()

    @Inputs.annotation_set
    def set_annotation_set(self, annotation_set: AnnotationSet | None) -> None:
        """Receive annotations used to constrain spool selection by overlap."""
        self._annotation_set = annotation_set
        self._emit_selected_output()

    def _selection_on_state_changed(self) -> None:
        """Recompute the current selected output and preview."""
        self._persist_selection_settings()
        self._emit_selected_output()

    def _emit_selected_output(self) -> None:
        """Emit the selected object on the output channel matching the input kind."""
        self.Error.clear()
        if self._input_kind == "patch":
            if self._patch is None:
                self._preview_selected = None
                self.Outputs.spool.send(None)
                self.Outputs.patch.send(None)
                self._request_ui_refresh()
                return
            try:
                selected = self._selection_apply_to_patch(self._patch)
            except Exception as exc:
                self._show_exception("general", exc)
                selected = self._patch
            self._preview_selected = selected
            self.Outputs.spool.send(None)
            self.Outputs.patch.send(selected)
            self._request_ui_refresh()
            return

        if self._input_kind == "spool":
            if self._spool is None:
                self._preview_selected = None
                self.Outputs.spool.send(None)
                self.Outputs.patch.send(None)
                self._request_ui_refresh()
                return
            try:
                selected = self._apply_annotation_filter_to_spool(self._spool)
                selected = self._selection_apply_to_spool(selected)
            except Exception as exc:
                self._show_exception("general", exc)
                selected = self._spool
            self._preview_selected = selected
            self.Outputs.spool.send(selected)
            self.Outputs.patch.send(self._extract_output_patch(selected))
            self._request_ui_refresh()
            return

        self._preview_selected = None
        self.Outputs.spool.send(None)
        self.Outputs.patch.send(None)
        self._request_ui_refresh()

    def _apply_annotation_filter_to_spool(self, spool: dc.BaseSpool) -> dc.BaseSpool:
        """Return only spool rows whose contents overlap the current annotations."""
        annotation_set = self._annotation_set
        if annotation_set is None:
            return spool
        contents = spool.get_contents()
        filtered = filter_contents_by_annotations(contents, annotation_set)
        if len(filtered) == len(contents):
            return spool
        if filtered.empty:
            return dc.spool([])
        wanted_rows = set(map(int, filtered.index))
        return dc.spool(
            [patch for row, patch in enumerate(spool) if row in wanted_rows]
        )

    def _refresh_ui(self) -> None:
        """Refresh the left-side selection controls and status text."""
        self._selection_refresh_panel()
        self.unpack_checkbox.setVisible(self._input_kind == "spool")
        self.unpack_checkbox.setEnabled(self._input_kind == "spool")
        self._selection_set_status(
            self._build_status_text(selected=self._preview_selected)
        )

    def _extract_output_patch(self, spool: dc.BaseSpool | None) -> dc.Patch | None:
        """Return the patch output when the emitted spool has length one."""
        if not self.unpack_single_patch or spool is None:
            return None
        return extract_single_patch(spool)

    def _load_saved_patch_selection_state(self) -> dict[str, object] | None:
        """Return serialized patch selection settings staged from widget state."""
        basis_name = str(self.saved_selection_basis or "").strip()
        rows = (
            self.saved_selection_ranges
            if isinstance(self.saved_selection_ranges, list)
            else []
        )
        if not basis_name or not rows:
            return None
        return {"basis": basis_name, "rows": rows}

    def _load_saved_spool_filter_state(self) -> list[tuple[str, str]]:
        """Return persisted spool filter rows from widget settings."""
        rows = (
            self.saved_spool_filters
            if isinstance(self.saved_spool_filters, list)
            else []
        )
        restored: list[tuple[str, str]] = []
        for row in rows:
            if not isinstance(row, dict):
                continue
            key = str(row.get("key", "")).strip()
            raw_value = str(row.get("raw_value", ""))
            if not key and not raw_value:
                continue
            restored.append((key, raw_value))
        return restored

    def _prime_saved_selection_state(self) -> None:
        """Stage workflow-backed selection settings before any input arrives."""
        if self._selection_state.prime_patch_state_from_settings(
            self._load_saved_patch_selection_state()
        ):
            self._selection_state.mode = SelectionMode.NONE
        saved_filters = self._load_saved_spool_filter_state()
        if saved_filters:
            self._selection_state.set_spool_filters(saved_filters)

    def _persist_selection_settings(self) -> None:
        """Mirror the current selection controls into schema-backed settings."""
        if self._input_kind == "patch":
            payload = self._selection_state.patch_settings_payload(
                include_inactive=True
            )
            if payload is None:
                self.saved_selection_basis = ""
                self.saved_selection_ranges = []
            else:
                self.saved_selection_basis = str(payload["basis"])
                self.saved_selection_ranges = list(payload["rows"])
            return

        if self._input_kind == "spool":
            self.saved_spool_filters = [
                {"key": row.key, "raw_value": row.raw_value}
                for row in self._selection_state.spool.filters
                if row.key or row.raw_value
            ]

    def _build_status_text(self, selected=None) -> str:
        """Return one compact status line for the current selection state."""
        if self._input_kind == "patch":
            if self._patch is None:
                return ""
            active = self._selection_state.patch_kwargs()
            basis = self._selection_state.patch.basis.value
            return f"{basis} basis, {len(active)} active range filter(s)"

        if self._input_kind == "spool":
            if self._spool is None:
                return ""
            source_count = len(self._spool.get_contents())
            result_count = (
                len(selected.get_contents()) if selected is not None else source_count
            )
            return f"{result_count} of {source_count} spool item(s) selected"

        return ""

    def _compact_initial_width(self) -> None:
        """Shrink one newly shown Select window to the width its controls need."""
        if self._compact_width_done or not self._is_ui_visible():
            return
        window = self.window()
        if window.isFullScreen() or window.isMaximized():
            self._compact_width_done = True
            return
        target_width = max(window.minimumWidth(), self.sizeHint().width())
        if window.width() > target_width:
            window.resize(target_width, window.height())
        self._compact_width_done = True


if __name__ == "__main__":  # pragma: no cover
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    WidgetPreview(Select).run()



"""Runtime helpers and widget classes for grouped composite workflows."""

from __future__ import annotations

import io
import uuid
from copy import deepcopy
from typing import Any

from AnyQt.QtWidgets import QApplication, QLabel
from Orange.widgets.widget import OWWidget
from orangecanvas.registry import InputSignal, OutputSignal, WidgetDescription
from orangecanvas.scheme import readwrite

from derzug.orange import Setting

COMPOSITE_PAYLOAD_KEY = "composite_payload"
NODE_ID_KEY = "__derzug_node_id"
INTERNAL_NODE_ID_KEY = "__derzug_composite_internal_node_id"
COMPOSITE_QNAME_PREFIX = "derzug.widgets.composite.DynamicComposite_"
_COMPOSITE_ICON = "icons/FolderGroup.svg"


def is_composite_qualified_name(qualified_name: str) -> bool:
    """Return True when one qualified name belongs to a dynamic composite."""
    return str(qualified_name).startswith(COMPOSITE_QNAME_PREFIX)


def ensure_node_id(node) -> str:
    """Return a stable DerZug node id, creating it when missing."""
    properties = dict(node.properties or {})
    node_id = str(properties.get(NODE_ID_KEY, "")).strip()
    if node_id:
        return node_id
    node_id = uuid.uuid4().hex
    properties[NODE_ID_KEY] = node_id
    node.properties = properties
    return node_id


def get_node_id(node) -> str | None:
    """Return the persisted DerZug node id when available."""
    properties = node.properties or {}
    value = properties.get(NODE_ID_KEY)
    return str(value).strip() or None if value is not None else None


def get_internal_node_id(node) -> str | None:
    """Return the composite-internal node id when available."""
    properties = node.properties or {}
    value = properties.get(INTERNAL_NODE_ID_KEY)
    return str(value).strip() or None if value is not None else None


def composite_payload_from_properties(properties: object) -> dict[str, Any] | None:
    """Return the composite payload from a properties dict when present."""
    if not isinstance(properties, dict):
        return None
    payload = properties.get(COMPOSITE_PAYLOAD_KEY)
    return payload if isinstance(payload, dict) else None


def composite_properties(
    payload: dict[str, Any], *, node_id: str | None = None
) -> dict[str, Any]:
    """Return initial node properties for one composite node."""
    properties = {COMPOSITE_PAYLOAD_KEY: deepcopy(payload)}
    properties[NODE_ID_KEY] = node_id or uuid.uuid4().hex
    return properties


def _clone_input_signal(
    signal: InputSignal,
    *,
    name: str,
    handler: str,
) -> InputSignal:
    """Return one input signal cloned with a new public name/handler."""
    return InputSignal(
        name=name,
        type=signal.type,
        handler=handler,
        flags=signal.flags,
        id=signal.id,
        doc=signal.doc,
        replaces=signal.replaces,
    )


def _clone_output_signal(signal: OutputSignal, *, name: str) -> OutputSignal:
    """Return one output signal cloned with a new public name."""
    return OutputSignal(
        name=name,
        type=signal.type,
        flags=signal.flags,
        id=signal.id,
        doc=signal.doc,
        replaces=signal.replaces,
    )


def _sanitize_token(text: str) -> str:
    """Return one ASCII-ish identifier fragment for generated class names."""
    return "".join(ch if ch.isalnum() else "_" for ch in str(text)).strip("_") or "x"


def _composite_class_name(payload: dict[str, Any]) -> str:
    """Return the generated composite widget class name for one payload."""
    composite_id = str(payload["composite_id"])
    return f"DynamicComposite_{_sanitize_token(composite_id)}"


def _bridge_input_class_name(payload: dict[str, Any]) -> str:
    """Return the generated internal input-bridge class name."""
    composite_id = str(payload["composite_id"])
    return f"DynamicCompositeInputBridge_{_sanitize_token(composite_id)}"


def _bridge_output_class_name(payload: dict[str, Any]) -> str:
    """Return the generated internal output-bridge class name."""
    composite_id = str(payload["composite_id"])
    return f"DynamicCompositeOutputBridge_{_sanitize_token(composite_id)}"


def composite_widget_description(payload: dict[str, Any]) -> WidgetDescription:
    """Return the runtime widget description for one composite payload."""
    qname = payload.get("qualified_name") or (
        f"{COMPOSITE_QNAME_PREFIX}{payload['composite_id']}"
    )
    payload["qualified_name"] = qname
    klass = ensure_composite_widget_class(payload)
    return WidgetDescription(
        name=str(payload.get("display_name") or "Composite"),
        id=f"derzug.dynamic-composite.{payload['composite_id']}",
        category="Processing",
        description="Composite workflow widget",
        qualified_name=f"{klass.__module__}.{klass.__name__}",
        package="derzug",
        project_name="derzug",
        icon=_COMPOSITE_ICON,
        inputs=list(klass.inputs),
        outputs=list(klass.outputs),
    )


class _CompositeInputBridgeBase(OWWidget, openclass=True):
    """Base class for composite runtime input bridges."""

    name = "Composite Input Bridge"
    want_main_area = False
    want_control_area = False
    resizing_enabled = False

    def __init__(self) -> None:
        super().__init__()


class _CompositeOutputBridgeBase(OWWidget, openclass=True):
    """Base class for composite runtime output bridges."""

    name = "Composite Output Bridge"
    want_main_area = False
    want_control_area = False
    resizing_enabled = False

    def __init__(self) -> None:
        super().__init__()
        self._composite = None

    def attach_composite(self, composite) -> None:
        """Attach the owning composite widget for forwarded outputs."""
        self._composite = composite


class _BaseCompositeWidget(OWWidget, openclass=True):
    """Base class for runtime composite widgets."""

    name = "Composite"
    description = "Composite workflow widget"
    icon = _COMPOSITE_ICON
    category = "Processing"
    want_main_area = False
    resizing_enabled = False

    composite_payload = Setting({})

    def __init__(self) -> None:
        super().__init__()
        self._runtime_scheme = None
        self._input_bridge = None
        self._output_bridge = None
        layout = self.controlArea.layout()
        summary = str(self.composite_payload.get("summary") or "Composite widget")
        self._summary_label = QLabel(summary, self.controlArea)
        self._summary_label.setWordWrap(True)
        layout.addWidget(self._summary_label)
        layout.addStretch(1)
        self._initialize_runtime()

    def _initialize_runtime(self) -> None:
        """Build the internal hidden workflow runtime for this composite."""
        payload = self.composite_payload or {}
        if not payload:
            return
        registry = _runtime_registry()
        if registry is None:
            return

        from derzug.views.orange import DerZugWidgetsScheme

        workflow = DerZugWidgetsScheme()
        xml_text = str(payload.get("internal_scheme_xml") or "")
        if xml_text:
            readwrite.scheme_load(
                workflow,
                io.BytesIO(xml_text.encode("utf-8")),
                registry=registry,
            )

        input_bridge_desc = _bridge_input_description(payload)
        output_bridge_desc = _bridge_output_description(payload)
        input_bridge_node = workflow.new_node(
            input_bridge_desc,
            title="Composite Inputs",
            position=(-240.0, 0.0),
        )
        output_bridge_node = workflow.new_node(
            output_bridge_desc,
            title="Composite Outputs",
            position=(240.0, 0.0),
        )

        internal_nodes_by_id = {
            get_internal_node_id(node): node
            for node in workflow.nodes
            if get_internal_node_id(node) is not None
        }

        for spec in payload.get("input_specs", []):
            target = internal_nodes_by_id.get(spec["internal_node_id"])
            if target is None:
                continue
            workflow.new_link(
                input_bridge_node,
                input_bridge_node.output_channel(spec["port_name"]),
                target,
                target.input_channel(spec["internal_channel_name"]),
            )

        for spec in payload.get("output_specs", []):
            source = internal_nodes_by_id.get(spec["internal_node_id"])
            if source is None:
                continue
            workflow.new_link(
                source,
                source.output_channel(spec["internal_channel_name"]),
                output_bridge_node,
                output_bridge_node.input_channel(spec["port_name"]),
            )

        for node in tuple(workflow.nodes):
            workflow.widget_manager.widget_for_node(node)

        self._runtime_scheme = workflow
        self._input_bridge = workflow.widget_manager.widget_for_node(input_bridge_node)
        self._output_bridge = workflow.widget_manager.widget_for_node(
            output_bridge_node
        )
        if self._output_bridge is not None:
            self._output_bridge.attach_composite(self)

    def _emit_internal_input(self, port_name: str, value: Any) -> None:
        """Push one external composite input into the hidden runtime."""
        if self._input_bridge is None:
            return
        handler = getattr(
            self._input_bridge,
            f"emit_{_sanitize_token(port_name)}",
            None,
        )
        if handler is None:
            return
        handler(value)

    def _forward_internal_output(self, port_name: str, value: Any) -> None:
        """Emit one exposed internal output to the outer Orange workflow."""
        self.send(port_name, value)


def _runtime_registry():
    """Return the live DerZug widget registry when available."""
    from orangecanvas.registry import global_registry

    try:
        import derzug.views.orange as orange_module
    except Exception:
        derzug_main_window_cls = None
    else:
        derzug_main_window_cls = orange_module.DerZugMainWindow
    if derzug_main_window_cls is not None:
        for widget in QApplication.topLevelWidgets():
            if isinstance(widget, derzug_main_window_cls):
                return widget.widget_registry
    return global_registry()


def ensure_composite_widget_class(payload: dict[str, Any]):
    """Return the generated composite widget class for one payload."""
    class_name = _composite_class_name(payload)
    existing = globals().get(class_name)
    if existing is not None:
        return existing

    input_specs = payload.get("input_specs", [])
    output_specs = payload.get("output_specs", [])
    attrs: dict[str, Any] = {
        "__module__": __name__,
        "name": str(payload.get("display_name") or "Composite"),
        "description": "Composite workflow widget",
        "icon": _COMPOSITE_ICON,
        "category": "Composite",
        "inputs": [
            _clone_input_signal(
                spec["signal"],
                name=spec["port_name"],
                handler=f"set_{_sanitize_token(spec['port_name'])}",
            )
            for spec in input_specs
        ],
        "outputs": [
            _clone_output_signal(spec["signal"], name=spec["port_name"])
            for spec in output_specs
        ],
    }

    for spec in input_specs:
        port_name = spec["port_name"]

        def _handler(self, value, *, _port_name=port_name):
            self._emit_internal_input(_port_name, value)

        attrs[f"set_{_sanitize_token(port_name)}"] = _handler

    klass = type(class_name, (_BaseCompositeWidget,), attrs)
    globals()[class_name] = klass
    ensure_bridge_input_class(payload)
    ensure_bridge_output_class(payload)
    return klass


def ensure_bridge_input_class(payload: dict[str, Any]):
    """Return the generated internal input-bridge class for one payload."""
    class_name = _bridge_input_class_name(payload)
    existing = globals().get(class_name)
    if existing is not None:
        return existing

    output_specs = payload.get("input_specs", [])
    attrs: dict[str, Any] = {
        "__module__": __name__,
        "outputs": [
            _clone_output_signal(spec["signal"], name=spec["port_name"])
            for spec in output_specs
        ],
        "inputs": [],
    }
    for spec in output_specs:
        port_name = spec["port_name"]

        def _emit(self, value, *, _port_name=port_name):
            self.send(_port_name, value)

        attrs[f"emit_{_sanitize_token(port_name)}"] = _emit

    klass = type(class_name, (_CompositeInputBridgeBase,), attrs)
    globals()[class_name] = klass
    return klass


def ensure_bridge_output_class(payload: dict[str, Any]):
    """Return the generated internal output-bridge class for one payload."""
    class_name = _bridge_output_class_name(payload)
    existing = globals().get(class_name)
    if existing is not None:
        return existing

    input_specs = payload.get("output_specs", [])
    attrs: dict[str, Any] = {
        "__module__": __name__,
        "inputs": [
            _clone_input_signal(
                spec["signal"],
                name=spec["port_name"],
                handler=f"set_{_sanitize_token(spec['port_name'])}",
            )
            for spec in input_specs
        ],
        "outputs": [],
    }
    for spec in input_specs:
        port_name = spec["port_name"]

        def _handler(self, value, *, _port_name=port_name):
            if self._composite is not None:
                self._composite._forward_internal_output(_port_name, value)

        attrs[f"set_{_sanitize_token(port_name)}"] = _handler

    klass = type(class_name, (_CompositeOutputBridgeBase,), attrs)
    globals()[class_name] = klass
    return klass


def _bridge_input_description(payload: dict[str, Any]) -> WidgetDescription:
    """Return the synthetic description for the internal input bridge."""
    klass = ensure_bridge_input_class(payload)
    return WidgetDescription(
        name="Composite Inputs",
        id=f"derzug.dynamic-composite-input.{payload['composite_id']}",
        category="Processing",
        qualified_name=f"{klass.__module__}.{klass.__name__}",
        package="derzug",
        project_name="derzug",
        outputs=list(klass.outputs),
        icon=_COMPOSITE_ICON,
    )


def _bridge_output_description(payload: dict[str, Any]) -> WidgetDescription:
    """Return the synthetic description for the internal output bridge."""
    klass = ensure_bridge_output_class(payload)
    return WidgetDescription(
        name="Composite Outputs",
        id=f"derzug.dynamic-composite-output.{payload['composite_id']}",
        category="Processing",
        qualified_name=f"{klass.__module__}.{klass.__name__}",
        package="derzug",
        project_name="derzug",
        inputs=list(klass.inputs),
        icon=_COMPOSITE_ICON,
    )



"""
Testing utilities for DerZug.

Provides helpers for instantiating Orange widgets.
"""

from __future__ import annotations

import os
import time
from contextlib import contextmanager
from dataclasses import dataclass
from typing import ClassVar, TypeVar

import pytest
from AnyQt.QtCore import Qt
from AnyQt.QtTest import QTest
from AnyQt.QtWidgets import QApplication
from Orange.widgets.tests.base import (
    DummySignalManager,
    WidgetTest,
    open_widget_classes,
)
from Orange.widgets.widget import OWWidget
from orangecanvas.registry import WidgetRegistry

from derzug.core import ZugWidget
from derzug.core.patchdimwidget import PatchDimWidget

T = TypeVar("T", bound=OWWidget)


def wait_for_widget_idle(widget: OWWidget, timeout: float = 2.0) -> None:
    """Pump Qt events until a widget finishes all active background tasks.

    Handles widgets that chain tasks (e.g. spool-load followed by a deferred
    emit task started via QTimer.singleShot): after each task completes, one
    extra processEvents() call flushes any pending singleShot(0, ...) timers
    that might start a follow-on task before declaring the widget idle.
    """
    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        QApplication.processEvents()
        if getattr(widget, "task", None) is None:
            # Double-check is intentional: Spool chains two tasks (load-index
            # then read-patch) via QTimer.singleShot(0, ...).  The first
            # processEvents() above fires on_done for task-1 and queues the
            # singleShot callback; this second call executes that callback and
            # starts task-2.  Without it we'd declare the widget idle while
            # task-2 is still pending, causing races in tests.
            QApplication.processEvents()
            if getattr(widget, "task", None) is None:
                return


def capture_output(output_slot, monkeypatch) -> list:
    """Intercept sends on an Orange output slot; return the collected values list."""
    received: list = []
    monkeypatch.setattr(output_slot, "send", received.append)
    return received


def wait_for_output(qtbot, received: list, count: int = 1, timeout: int = 3000) -> None:
    """Block until `received` holds at least `count` values."""
    qtbot.waitUntil(lambda: len(received) >= count, timeout=timeout)


def wait_for_window_close(window: OWWidget, poll_interval: float = 0.01) -> None:
    """Keep processing events until the user closes an interactive test window."""
    while window.isVisible():
        QApplication.processEvents()
        time.sleep(poll_interval)


@dataclass
class BuiltWorkflow:
    """Container for a test workflow built in an Orange scheme."""

    scheme: object
    nodes_by_title: dict[str, object]
    widgets_by_title: dict[str, OWWidget]
    descriptions_by_name: dict[str, object]


@contextmanager
def widget_context(cls: type[T], stored_settings: dict | None = None):
    """
    Context manager for safely creating and cleaning up an OWidget in tests.

    Replicates the essential parts of Orange's WidgetTest.create_widget
    without requiring unittest-style test class inheritance.

    Parameters
    ----------
    cls : type[T]
        The OWidget subclass to instantiate.
    stored_settings : dict, optional
        Settings to restore on the widget, passed directly to __new__.

    Yields
    ------
    T
        The instantiated widget, ready for assertions.

    Examples
    --------
    >>> with widget_context(MyWidget) as w:
    ...     assert w is not None
    """
    signal_manager = DummySignalManager()

    # Rebind the settings handler and clear any defaults cached from disk so
    # the widget always starts from a known state, matching what
    # WidgetTest.create_widget does via reset_default_settings.
    handler = getattr(cls, "settingsHandler", None)
    if handler is not None:
        handler.bind(cls)
        handler.defaults = {}
        handler.global_contexts = []

    # open_widget_classes suppresses Orange's subclassing deprecation guard,
    # which fires when __new__ creates the internal tracking subclass.
    with open_widget_classes():
        widget = cls.__new__(
            cls,
            signal_manager=signal_manager,
            stored_settings=stored_settings,
        )
        widget.__init__()

    # Flush any singleShot timers queued during __init__.
    QApplication.processEvents()
    wait_for_widget_idle(widget)

    try:
        yield widget
    finally:
        # Mirror WidgetTest teardown: notify the widget, close, then schedule
        # deletion so Qt can reclaim resources on the next event loop tick.
        widget.onDeleteWidget()
        widget.close()
        widget.deleteLater()
        QApplication.processEvents()


def build_window_workflow(
    window,
    registry: WidgetRegistry,
    widgets: tuple[tuple[str, str], ...],
    links: tuple[tuple[str, str, str, str], ...] = (),
    *,
    qapp=None,
    clear: bool = True,
) -> BuiltWorkflow:
    """
    Build a workflow in the current window scheme from widget and link specs.

    Parameters
    ----------
    window
        The Orange main window whose current scheme should be populated.
    registry
        The widget registry used to resolve widget descriptions by name.
    widgets
        Tuples of ``(registry_name, title)`` for each node to create.
    links
        Tuples of ``(source_title, output_name, sink_title, input_name)`` for links.
    qapp
        Optional QApplication used to flush events after building.
    clear
        If True, clear all existing nodes from the current scheme before building.

    Returns
    -------
    BuiltWorkflow
        A container with the scheme, created nodes, widgets, and descriptions.
    """
    scheme = window.current_document().scheme()
    if clear:
        for node in list(scheme.nodes):
            scheme.remove_node(node)

    descriptions_by_name = {widget.name: widget for widget in registry.widgets()}
    nodes_by_title: dict[str, object] = {}

    for registry_name, title in widgets:
        desc = descriptions_by_name[registry_name]
        nodes_by_title[title] = scheme.new_node(desc, title=title)

    for source_title, output_name, sink_title, input_name in links:
        source_node = nodes_by_title[source_title]
        sink_node = nodes_by_title[sink_title]
        scheme.new_link(
            source_node,
            source_node.output_channel(output_name),
            sink_node,
            sink_node.input_channel(input_name),
        )

    if qapp is not None:
        qapp.processEvents()

    widgets_by_title = {
        title: scheme.widget_for_node(node) for title, node in nodes_by_title.items()
    }

    if qapp is not None:
        qapp.processEvents()

    return BuiltWorkflow(
        scheme=scheme,
        nodes_by_title=nodes_by_title,
        widgets_by_title=widgets_by_title,
        descriptions_by_name=descriptions_by_name,
    )


class TestWidgetDefaults(WidgetTest):
    """
    DerZug's default widget test suite.
    """

    __test__ = False

    widget: ClassVar[type[ZugWidget]]

    # The inputs needed to init the test Widget of the form:
    # ((input_name, input_value), ...)
    inputs: ClassVar[tuple[tuple[str, object], ...]] = ()
    stored_settings: ClassVar[dict | None] = None

    def create_default_widget(self):
        """Create the configured widget and apply any default input signals."""
        widget = self.create_widget(self.widget, stored_settings=self.stored_settings)
        for input_name, value in self.inputs:
            sig = getattr(widget.Inputs, input_name)
            self.send_signal(sig, value, widget=widget)
        widget.show()
        self.process_events()
        wait_for_widget_idle(widget)
        return widget

    @pytest.fixture()
    def widget_object(self):
        """Yield the default widget instance for tests that need a live widget."""
        widget = self.create_default_widget()
        yield widget
        self.process_events()

    def test_widget_instantiates(self):
        """Widget instantiates with shared metadata and signal definitions."""
        widget_object = self.create_default_widget()
        assert isinstance(widget_object, self.widget)
        assert isinstance(widget_object, ZugWidget)
        assert bool(getattr(widget_object, "name", "").strip())
        assert bool(getattr(widget_object, "description", "").strip())
        assert bool(getattr(widget_object, "category", "").strip())
        assert hasattr(widget_object, "Inputs")
        assert hasattr(widget_object, "Outputs")
        error_general = getattr(widget_object.Error, "general", None)
        if error_general is not None:
            assert not error_general.is_shown()

    def test_minimum_size(self):
        """Widget meets Orange minimum-size expectations."""
        self.check_minimum_size(self.create_default_widget())

    def test_image_export(self):
        """Widget supports Orange's image export smoke check."""
        self.check_export_image(self.create_default_widget())

    def test_msg_base_class(self):
        """Widget message groups inherit from the expected base class."""
        self.check_msg_base_class(self.create_default_widget())

    def test_control_area_items_are_top_aligned(self):
        """Control-area content stays packed toward the top of the sidebar."""
        widget_object = self.create_default_widget()
        layout = widget_object.controlArea.layout()

        assert layout is not None
        assert bool(layout.alignment() & Qt.AlignTop)

    def test_ctrl_q_closes_window(self):
        """Pressing Ctrl+Q closes the shown widget window."""
        widget_object = self.create_default_widget()
        widget_object.show()
        self.process_events()
        window = widget_object.window()
        window.activateWindow()
        window.raise_()
        self.process_events()
        assert window.isVisible()

        QTest.keyClick(window, Qt.Key_Q, Qt.ControlModifier)
        deadline = time.monotonic() + 1.0
        while window.isVisible() and time.monotonic() < deadline:
            self.process_events()
            time.sleep(0.01)
        assert not window.isVisible()

    def test_f_enters_fullscreen(self):
        """Pressing f makes main-area widgets enter fullscreen."""
        widget_object = self.create_default_widget()
        if not bool(getattr(widget_object, "want_main_area", True)):
            pytest.skip("Widget does not use a main area")

        widget_object.show()
        self.process_events()
        window = widget_object.window()
        window.activateWindow()
        window.raise_()
        self.process_events()
        assert not window.isFullScreen()

        QTest.keyClick(window, Qt.Key_F)
        deadline = time.monotonic() + 1.0
        while not window.isFullScreen() and time.monotonic() < deadline:
            self.process_events()
            time.sleep(0.01)
        assert window.isFullScreen()
        window.showNormal()
        self.process_events()

    @pytest.mark.show
    def test_show_widget_window(self):
        """Show the widget window and keep it open for manual interaction."""
        widget_object = self.create_default_widget()
        self.show(widget_object)
        window = widget_object.window()
        window.activateWindow()
        window.raise_()
        self.process_events()

        # Interactive show-mode runs are intentionally blocking so a developer
        # can inspect and manipulate the live widget before closing it.
        if os.environ.get("QT_QPA_PLATFORM") != "offscreen":
            wait_for_window_close(window)


class TestPatchDimWidgetDefaults(TestWidgetDefaults):
    """Shared defaults for PatchDimWidget subclasses with selected_dim state."""

    __test__ = False

    widget: ClassVar[type[PatchDimWidget]]
    compatible_patch: ClassVar[object]
    incompatible_patch: ClassVar[object]
    persisted_dim: ClassVar[str] = "time"

    def test_selected_dim_survives_none_then_compatible_patch(self):
        """`None` should not clear the stored dim before a compatible replacement."""
        widget_object = self.create_default_widget()
        assert isinstance(widget_object, PatchDimWidget)
        signal = getattr(widget_object.Inputs, "patch")

        self.send_signal(signal, self.compatible_patch, widget=widget_object)
        self.process_events()
        widget_object.selected_dim = self.persisted_dim

        self.send_signal(signal, None, widget=widget_object)
        self.process_events()
        assert widget_object.selected_dim == self.persisted_dim

        self.send_signal(signal, self.compatible_patch, widget=widget_object)
        self.process_events()
        assert widget_object.selected_dim == self.persisted_dim

    def test_selected_dim_resets_only_on_incompatible_replacement(self):
        """Stored dims should reset only when the next real patch cannot use them."""
        widget_object = self.create_default_widget()
        assert isinstance(widget_object, PatchDimWidget)
        signal = getattr(widget_object.Inputs, "patch")

        self.send_signal(signal, self.compatible_patch, widget=widget_object)
        self.process_events()
        widget_object.selected_dim = self.persisted_dim

        self.send_signal(signal, None, widget=widget_object)
        self.process_events()
        assert widget_object.selected_dim == self.persisted_dim

        self.send_signal(signal, self.incompatible_patch, widget=widget_object)
        self.process_events()
        assert widget_object.selected_dim in getattr(
            widget_object, "_available_dims", ()
        )
        assert widget_object.selected_dim != self.persisted_dim


class TestPatchInputStateDefaults(TestWidgetDefaults):
    """Shared defaults for patch widgets with custom persisted input state."""

    __test__ = False

    compatible_patch: ClassVar[object]
    incompatible_patch: ClassVar[object]
    input_signal_name: ClassVar[str] = "patch"

    def _resolve_input_value(self, value):
        """Materialize deferred test inputs when needed."""
        return value() if callable(value) else value

    def arrange_persisted_input_state(self, widget_object):
        """Install persisted widget state after a compatible patch arrives."""
        raise NotImplementedError

    def assert_persisted_input_state(self, widget_object, state_token) -> None:
        """Assert the arranged state survives a compatible replacement."""
        raise NotImplementedError

    def assert_reset_input_state(self, widget_object, state_token) -> None:
        """Assert incompatible replacement input resets invalid stored state."""
        raise NotImplementedError

    def test_input_state_survives_none_then_compatible_patch(self):
        """`None` should not clear persisted state before a compatible replacement."""
        widget_object = self.create_default_widget()
        signal = getattr(widget_object.Inputs, self.input_signal_name)
        compatible = self._resolve_input_value(self.compatible_patch)

        self.send_signal(signal, compatible, widget=widget_object)
        self.process_events()
        state_token = self.arrange_persisted_input_state(widget_object)
        self.process_events()

        self.send_signal(signal, None, widget=widget_object)
        self.process_events()
        self.assert_persisted_input_state(widget_object, state_token)

        self.send_signal(
            signal,
            self._resolve_input_value(self.compatible_patch),
            widget=widget_object,
        )
        self.process_events()
        self.assert_persisted_input_state(widget_object, state_token)

    def test_input_state_resets_only_on_incompatible_replacement(self):
        """Persisted state should reset only when a real replacement cannot use it."""
        widget_object = self.create_default_widget()
        signal = getattr(widget_object.Inputs, self.input_signal_name)
        compatible = self._resolve_input_value(self.compatible_patch)

        self.send_signal(signal, compatible, widget=widget_object)
        self.process_events()
        state_token = self.arrange_persisted_input_state(widget_object)
        self.process_events()

        self.send_signal(signal, None, widget=widget_object)
        self.process_events()
        self.assert_persisted_input_state(widget_object, state_token)

        self.send_signal(
            signal,
            self._resolve_input_value(self.incompatible_patch),
            widget=widget_object,
        )
        self.process_events()
        self.assert_reset_input_state(widget_object, state_token)



"""Tests for pytest harness helpers."""

from __future__ import annotations

from conftest import _should_suppress_qt_message


class TestQtMessageFilter:
    """Tests for headless Qt message suppression."""

    def test_known_offscreen_plugin_noise_is_suppressed(self):
        """Known offscreen platform-plugin chatter should be filtered."""
        assert _should_suppress_qt_message(
            "This plugin does not support propagateSizeHints()"
        )
        assert _should_suppress_qt_message("This plugin does not support raise()")
        assert _should_suppress_qt_message(
            "This plugin does not support grabbing the keyboard"
        )

    def test_unrelated_qt_messages_are_not_suppressed(self):
        """Non-noise Qt messages should continue to surface."""
        assert not _should_suppress_qt_message("Real Qt warning")



"""Shared display formatting helpers for UI text."""

from __future__ import annotations

import datetime
from typing import Any

import numpy as np

from derzug.constants import DISPLAY_SIGFIGS


def _format_duration(seconds: float) -> str:
    """Format a duration as a concise, human-readable string."""
    abs_s = abs(seconds)
    if abs_s < 1e-6:
        return f"{seconds * 1e9:.3g} ns"
    if abs_s < 1e-3:
        return f"{seconds * 1e6:.3g} µs"
    if abs_s < 1.0:
        return f"{seconds * 1e3:.3g} ms"
    if abs_s < 60.0:
        return f"{seconds:.4g} s"
    m, s = divmod(seconds, 60.0)
    if abs_s < 3600.0:
        return f"{int(m)} m {s:.3g} s"
    h, rem = divmod(seconds, 3600.0)
    return f"{int(h)} h {int(rem / 60)} m"


def format_display(value: Any) -> str:
    """Return a stable UI display string for scalar values."""
    if value is None:
        return ""

    # Handle timedelta before np.asarray: pd.Timedelta is a datetime.timedelta
    # subclass but np.asarray returns object dtype in recent pandas versions.
    if isinstance(value, datetime.timedelta):
        return _format_duration(value.total_seconds())

    arr = np.asarray(value)
    if np.issubdtype(arr.dtype, np.datetime64):
        return np.datetime_as_string(arr.astype("datetime64[ns]"), timezone="naive")
    if np.issubdtype(arr.dtype, np.timedelta64):
        seconds = float(arr.astype("timedelta64[ns]").astype(np.int64)) / 1e9
        return _format_duration(seconds)

    if isinstance(value, np.generic):
        value = value.item()

    if isinstance(value, float):
        return f"{value:.{DISPLAY_SIGFIGS}g}"
    if isinstance(value, int):
        return str(value)
    return str(value)



"""Orange widget for DASCore Fourier-domain transforms."""

from __future__ import annotations

from typing import ClassVar

import dascore as dc
from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import (
    QCheckBox,
    QComboBox,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)
from Orange.widgets import gui
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.widget import Msg

from derzug.core.patchdimwidget import PatchDimWidget
from derzug.orange import Setting

_TRANSFORMS: tuple[str, ...] = ("dft", "idft")
_REAL_OPTIONS: tuple[tuple[str, str | bool | None], ...] = (
    ("Auto", None),
    ("Real", True),
    ("Complex", False),
)
_REAL_OPTION_MAP = dict(_REAL_OPTIONS)


class Fourier(PatchDimWidget):
    """Apply DASCore Fourier transforms to an input patch."""

    name = "Fourier"
    description = "Apply DASCore Fourier transforms to a patch"
    icon = "icons/Fourier.svg"
    category = "Transform"
    keywords = ("transform", "fourier", "fft", "dft", "idft")
    priority = 21.1
    want_main_area = False

    transform = Setting("dft")
    selected_dim = Setting("")
    selected_dims = Setting([])
    real_mode = Setting("Auto")
    pad = Setting(True)

    _TRANSFORMS: ClassVar[tuple[str, ...]] = _TRANSFORMS

    class Error(PatchDimWidget.Error):
        """Errors shown by the widget."""

        transform_failed = Msg("Fourier transform failed: {}")

    class Inputs:
        """Input signal definitions."""

        patch = Input("Patch", dc.Patch)

    class Outputs:
        """Output signal definitions."""

        patch = Output("Patch", dc.Patch)

    def __init__(self) -> None:
        super().__init__()
        box = gui.widgetBox(self.controlArea, "Parameters")
        gui.widgetLabel(box, "Transform:")
        self._transform_combo = QComboBox(box)
        self._transform_combo.addItems(self._TRANSFORMS)
        box.layout().addWidget(self._transform_combo)

        self._stack = QStackedWidget(box)
        box.layout().addWidget(self._stack)
        self._stack.addWidget(self._build_dft_page())
        self._stack.addWidget(self._build_idft_page())

        if self.transform not in self._TRANSFORMS:
            self.transform = self._TRANSFORMS[0]
        self._transform_combo.setCurrentText(self.transform)
        self._stack.setCurrentIndex(self._TRANSFORMS.index(self.transform))

        self._transform_combo.currentTextChanged.connect(self._on_transform_changed)
        self._dim_combo.currentTextChanged.connect(self._on_dim_changed)

    def _build_dft_page(self) -> QWidget:
        """Return the parameter page for forward DFT."""
        page = QWidget()
        layout = QVBoxLayout(page)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(QLabel("Dimensions:", page))
        self._dim_list = QListWidget(page)
        layout.addWidget(self._dim_list)

        layout.addWidget(QLabel("Real output:", page))
        self._real_combo = QComboBox(page)
        self._real_combo.addItems([label for label, _ in _REAL_OPTIONS])
        if self.real_mode not in _REAL_OPTION_MAP:
            self.real_mode = "Auto"
        self._real_combo.setCurrentText(self.real_mode)
        self._real_combo.currentTextChanged.connect(self._on_real_mode_changed)
        layout.addWidget(self._real_combo)

        self._pad_checkbox = QCheckBox("Pad input", page)
        self._pad_checkbox.setChecked(bool(self.pad))
        self._pad_checkbox.toggled.connect(self._on_pad_toggled)
        layout.addWidget(self._pad_checkbox)
        layout.addStretch(1)
        self._dim_list.itemChanged.connect(self._on_dim_list_changed)
        return page

    def _build_idft_page(self) -> QWidget:
        """Return the parameter page for inverse DFT."""
        page = QWidget()
        layout = QVBoxLayout(page)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(QLabel("Dimension:", page))
        self._dim_combo = QComboBox(page)
        layout.addWidget(self._dim_combo)
        layout.addWidget(QLabel("Inverse transform the selected Fourier axis.", page))
        layout.addStretch(1)
        return page

    @Inputs.patch
    def set_patch(self, patch: dc.Patch | None) -> None:
        """Receive an input patch and run the selected transform."""
        self._set_patch_input(patch)
        self.run()

    def _default_dim(self, dims: tuple[str, ...]) -> str:
        """Choose a default dimension, preferring common time-domain axes."""
        if "time" in dims:
            return "time"
        ft_dims = [dim for dim in dims if dim.startswith("ft_")]
        if ft_dims:
            return ft_dims[0]
        return dims[0]

    def _on_transform_changed(self, value: str) -> None:
        """Persist selected transform and rerun."""
        self.transform = value
        self._stack.setCurrentIndex(self._TRANSFORMS.index(value))
        self.run()

    def _on_dim_changed(self, value: str) -> None:
        """Persist selected dimension and rerun."""
        self.selected_dim = value
        self.run()

    def _on_dim_list_changed(self, _item: QListWidgetItem) -> None:
        """Persist selected DFT dimensions and rerun."""
        self.selected_dims = self._checked_dims()
        self.run()

    def _on_real_mode_changed(self, value: str) -> None:
        """Persist selected DFT real-mode option and rerun."""
        self.real_mode = value
        self.run()

    def _on_pad_toggled(self, value: bool) -> None:
        """Persist the DFT pad option and rerun."""
        self.pad = bool(value)
        self.run()

    def _coerce_transform(self) -> str:
        """Return the selected transform or reset to the default."""
        if self.transform in self._TRANSFORMS:
            return self.transform
        self.transform = self._TRANSFORMS[0]
        self._transform_combo.blockSignals(True)
        self._transform_combo.setCurrentText(self.transform)
        self._transform_combo.blockSignals(False)
        self._stack.setCurrentIndex(0)
        return self.transform

    def _refresh_dims(self) -> None:
        """Sync both forward- and inverse-transform dimension selectors."""
        dims = tuple(self._patch.dims) if self._patch is not None else ()
        self._available_dims = dims

        valid_selected_dims = [dim for dim in self.selected_dims if dim in dims]
        if dims and not valid_selected_dims:
            valid_selected_dims = [self._default_dim(dims)]
        self.selected_dims = valid_selected_dims

        self._dim_list.blockSignals(True)
        self._dim_list.clear()
        for dim in dims:
            item = QListWidgetItem(dim, self._dim_list)
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(
                Qt.Checked if dim in self.selected_dims else Qt.Unchecked
            )
        self._dim_list.setEnabled(bool(dims))
        self._dim_list.blockSignals(False)

        self._dim_combo.blockSignals(True)
        self._dim_combo.clear()
        self._dim_combo.addItems(dims)
        if dims:
            if self.selected_dim not in dims:
                ft_dims = [dim for dim in dims if dim.startswith("ft_")]
                self.selected_dim = (
                    self._default_dim(tuple(ft_dims))
                    if ft_dims
                    else self._default_dim(dims)
                )
            self._dim_combo.setCurrentText(self.selected_dim)
        self._dim_combo.setEnabled(bool(dims))
        self._dim_combo.blockSignals(False)

    def _get_dims(self) -> tuple[str, ...] | None:
        """Return the selected forward-transform dimensions when available."""
        if not self._available_dims:
            return None
        dims = tuple(dim for dim in self.selected_dims if dim in self._available_dims)
        if not dims:
            dims = (self._default_dim(self._available_dims),)
            self.selected_dims = list(dims)
            self._refresh_dims()
        return dims

    def _checked_dims(self) -> list[str]:
        """Return the checked DFT dimensions in display order."""
        selected: list[str] = []
        for index in range(self._dim_list.count()):
            item = self._dim_list.item(index)
            if item.checkState() == Qt.Checked:
                selected.append(item.text())
        return selected

    def _run(self) -> dc.Patch | None:
        """Apply the selected Fourier transform and return the output patch."""
        if self._patch is None:
            return None

        transform = self._coerce_transform()
        try:
            if transform == "dft":
                dims = self._get_dims()
                if dims is None:
                    return None
                real_mode = (
                    self.real_mode if self.real_mode in _REAL_OPTION_MAP else "Auto"
                )
                if real_mode != self.real_mode:
                    self.real_mode = real_mode
                    self._real_combo.blockSignals(True)
                    self._real_combo.setCurrentText(real_mode)
                    self._real_combo.blockSignals(False)
                return self._patch.dft(
                    dims,
                    real=_REAL_OPTION_MAP[real_mode],
                    pad=bool(self.pad),
                )
            dim = self._get_dim()
            if dim is None:
                return None
            return self._patch.idft(dim)
        except Exception as exc:
            self._show_exception("transform_failed", exc)
            return None


if __name__ == "__main__":  # pragma: no cover
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    WidgetPreview(Fourier).run()



"""Command-line interface for derzug."""

from __future__ import annotations

from pathlib import Path

import typer

app = typer.Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command()
def main(
    workflow: Path | None = typer.Argument(
        None, help="Workflow file (.ows) to load on startup."
    ),
    demo: bool = typer.Option(
        False, "--demo", help="Open the example workflow browser on startup."
    ),
    dev: bool = typer.Option(
        False, "--dev", help="Enable development UI features such as hot reload."
    ),
    open_widgets: str = typer.Option(
        "",
        "--open-widgets",
        help="Comma-separated node indices to reopen on startup (used by hot reload).",
        hidden=True,
    ),
) -> None:
    """Run the derzug command-line interface."""
    from derzug.views.orange import DerZugMain, ensure_linux_desktop_entry

    argv = ["derzug"]
    if workflow is not None:
        argv.append(str(workflow))
    ensure_linux_desktop_entry()
    runner = DerZugMain()
    runner.show_demo = demo
    runner.dev_mode = dev
    runner.startup_workflow_path = None if workflow is None else str(workflow)
    runner.startup_open_widget_ids = [
        int(x) for x in open_widgets.split(",") if x.strip().isdigit()
    ]
    raise SystemExit(runner.run(argv))


if __name__ == "__main__":
    app()
